"""
Entry point for standalone brewblox_service.

This will create the app, and enable events.
"""

from brewblox_service import events, scheduler, service


def main():
    app = service.create_app(default_name='_service')

    # Event handling is optional, but requires the scheduler
    # Both should be enabled explicitly by service implementations
    scheduler.setup(app)
    events.setup(app)

    # Add all default endpoints
    service.furnish(app)

    # service.run() will start serving clients async
    service.run(app)


if __name__ == '__main__':
    main()



"""
Offers event publishing and subscription for service implementations.

Example use:

    from brewblox_service import scheduler, events

    scheduler.setup(app)
    events.setup(app)

    async def on_message(subscription, key, message):
        print(f'Message from {subscription}: {key} = {message} ({type(message)})')

    events.subscribe(app, 'brewblox', 'controller', on_message=on_message)
    events.subscribe(app, 'brewblox', 'controller.*', on_message=on_message)
    events.subscribe(app, 'brewblox', 'controller.#', on_message=on_message)

    await events.publish(app, 'brewblox', 'controller.value', {'example': True})

"""

import asyncio
import json
import queue
import warnings
from datetime import timedelta
from typing import Callable, Coroutine, List, Literal, Optional, Union

import aioamqp
from aiohttp import web

from brewblox_service import brewblox_logger, features, repeater, strex

LOGGER = brewblox_logger(__name__)
routes = web.RouteTableDef()

EvtCallbackType_ = Callable[['EventSubscription', str, Union[dict, str]], Coroutine]
ExchangeType_ = Literal['topic', 'fanout', 'direct']

RECONNECT_INTERVAL = timedelta(seconds=1)
PENDING_WAIT_TIMEOUT = timedelta(seconds=5)

##############################################################################
# Incoming events
##############################################################################


async def _default_on_message(sub: 'EventSubscription', key: str, message: Union[dict, str]):
    LOGGER.info(f'Unhandled event: subscription={sub}, key={key}, message={message}')


class EventSubscription():
    """
    Subscription class for receiving AMQP messages.

    This class should not be instantiated directly.
    To subscribe to AMQP messages, use `EventListener.subscribe()`

    The `on_message` property can safely be changed while the subscription is active.
    It will be used for the next received message.
    """

    def __init__(self,
                 exchange_name: str,
                 routing: str,
                 exchange_type: ExchangeType_ = 'topic',
                 on_message: Optional[EvtCallbackType_] = None):
        self._routing: str = routing
        self._exchange_name: str = exchange_name
        self._exchange_type: ExchangeType_ = exchange_type
        self._on_message: EvtCallbackType_ = on_message or _default_on_message

    def __str__(self):
        return f'<{self._routing} @ {self._exchange_name}>'

    @property
    def on_message(self) -> EvtCallbackType_:
        return self._on_message

    @on_message.setter
    def on_message(self, f: Optional[EvtCallbackType_]):
        self._on_message = f or _default_on_message

    async def declare_on_remote(self, channel: aioamqp.channel.Channel):
        LOGGER.info(f'Declaring event bus subscription {self} on {channel}')

        await channel.exchange_declare(
            exchange_name=self._exchange_name,
            type_name=self._exchange_type,
            auto_delete=True
        )

        queue_info = await channel.queue_declare(exclusive=True)
        queue_name = queue_info['queue']

        await channel.queue_bind(
            queue_name=queue_name,
            exchange_name=self._exchange_name,
            routing_key=self._routing
        )

        await channel.basic_consume(
            callback=self._relay,
            queue_name=queue_name
        )

    async def _relay(self,
                     channel: aioamqp.channel.Channel,
                     body: str,
                     envelope: aioamqp.envelope.Envelope,
                     properties: aioamqp.properties.Properties):
        """Relays incoming messages between the queue and the user callback"""
        try:
            await channel.basic_client_ack(envelope.delivery_tag)
            await self.on_message(self, envelope.routing_key, json.loads(body))
        except Exception as ex:
            LOGGER.error(f'Exception relaying message in {self}: {ex}')


class EventListener(repeater.RepeaterFeature):
    """
    Allows subscribing to AMQP messages published to a central event bus.

    `EventListener` will maintain a persistent connection to the AMQP host,
    and ensures that all subscriptions remain valid if the connection is lost and reestablished.
    """

    def __init__(self,
                 app: web.Application,
                 host: str = None,
                 port: int = None,
                 ):
        super().__init__(app)

        self._host: str = host or app['config']['eventbus_host']
        self._port: int = port or app['config']['eventbus_port']
        self._last_ok: bool = True

        # Asyncio queues need a context loop
        # We'll replace _pending with an asyncio.Queue when we can
        self._pending: Union[queue.Queue, asyncio.Queue] = queue.Queue()
        self._active: List[EventSubscription] = []

        # Same as the Queue: it will be initialized when async
        # The background task will await this before connecting
        self._has_pending: Optional[asyncio.Event] = None

    def __str__(self):
        return f'<{type(self).__name__} for {self._host}>'

    async def prepare(self):
        """
        Overrides RepeaterFeature.prepare().
        This function will be called once before the run() loop.
        """
        # Initialize async versions of objects
        # Don't discard the sync queue
        sync_pending = self._pending
        self._pending = asyncio.Queue()
        self._has_pending = asyncio.Event()

        # Migrate data from sync Queue to async Queue
        while sync_pending.qsize() > 0:
            self._pending.put_nowait(sync_pending.get_nowait())
            self._has_pending.set()

        await self._has_pending.wait()
        LOGGER.info(f'{self} now listening')

    async def run(self):
        """
        Overrides RepeaterFeature.run().
        This function will be called in a loop in a background Task.
        """
        transport: asyncio.Transport = None
        protocol: aioamqp.AmqpProtocol = None
        channel: aioamqp.channel.Channel = None

        try:
            transport, protocol = await aioamqp.connect(
                host=self._host,
                port=self._port,
                login_method='PLAIN',
            )
            channel = await protocol.channel()

            LOGGER.info(f'Connected {self}')

            # Declare all current subscriptions if reconnecting
            [await sub.declare_on_remote(channel) for sub in self._active]

            while True:
                subscription: EventSubscription = None

                try:
                    await protocol.ensure_open()
                    self._last_ok = True

                    subscription = await asyncio.wait_for(
                        self._pending.get(),
                        timeout=PENDING_WAIT_TIMEOUT.seconds
                    )

                except asyncio.TimeoutError:  # pragma: no cover
                    # Timeout ensures that connection state is checked at least once per timeout
                    continue

                try:
                    await protocol.ensure_open()
                    await subscription.declare_on_remote(channel)
                    self._active.append(subscription)

                except Exception:
                    # Put subscription back in queue
                    # We'll declare it after reconnect
                    self._pending.put_nowait(subscription)
                    raise

        except Exception as ex:
            if self._last_ok:
                warnings.warn(f'Connection error in {self}: {strex(ex)}')
                self._last_ok = False

            await asyncio.sleep(RECONNECT_INTERVAL.seconds)

        finally:
            protocol and await protocol.close()
            transport and transport.close()

    def subscribe(self,
                  exchange_name: str,
                  routing: str,
                  exchange_type: ExchangeType_ = 'topic',
                  on_message: EvtCallbackType_ = None
                  ) -> EventSubscription:
        """Adds a new event subscription to the listener.

        Actual queue declaration to the remote message server is done when connected.
        If the listener is not currently connected, it defers declaration.

        All existing subscriptions are redeclared on the remote if `EventListener`
        loses and recreates the connection.

        Args:
            exchange_name (str):
                Name of the AMQP exchange. Messages are always published to a specific exchange.

            routing (str):
                Filter messages passing through the exchange.
                A routing key is a '.'-separated string, and accepts '#' and '*' wildcards.

            exchange_type (ExchangeType_, optional):
                If the exchange does not yet exist, it will be created with this type.
                Default is `topic`, acceptable values are `topic`, `fanout`, or `direct`.

            on_message (EvtCallbackType_, optional):
                The function to be called when a new message is received.
                If `on_message` is none, it will default to logging the message.

        Returns:
            EventSubscription:
                The newly created subscription.
                This value can safely be discarded: EventListener keeps its own reference.
        """
        sub = EventSubscription(
            exchange_name,
            routing,
            exchange_type,
            on_message=on_message
        )
        self._pending.put_nowait(sub)

        if self._has_pending:
            self._has_pending.set()
        else:
            LOGGER.info(f'Deferred event bus subscription: [{sub}]')

        return sub

##############################################################################
# Outgoing events
##############################################################################


class EventPublisher(features.ServiceFeature):
    """
    Allows publishing AMQP messages to a central eventbus.

    `EventPublisher` is associated with a single eventbus address,
    but will only create a connection when attempting to publish.

    Connections are re-used for subsequent `publish()` calls.
    """

    def __init__(self,
                 app: web.Application,
                 host: str = None,
                 port: int = None
                 ):
        super().__init__(app)

        self._host: str = host or app['config']['eventbus_host']
        self._port: int = port or app['config']['eventbus_port']
        self._transport: asyncio.Transport
        self._protocol: aioamqp.AmqpProtocol
        self._channel: aioamqp.channel.Channel
        self._reset()

    @property
    def connected(self):
        return self._transport and self._protocol and self._channel

    def __str__(self):
        return f'<{type(self).__name__} for "{self._host}">'

    def _reset(self):
        self._transport = None
        self._protocol = None
        self._channel = None

    async def _close(self):
        LOGGER.info(f'Closing {self}')

        try:
            await self._protocol.close(no_wait=True)
            self._transport.close()
        except Exception:
            pass
        finally:
            self._reset()

    async def _ensure_channel(self):
        if not self.connected:
            self._transport, self._protocol = await aioamqp.connect(
                host=self._host,
                port=self._port,
                login_method='PLAIN',
            )
            self._channel = await self._protocol.channel()

        try:
            await self._protocol.ensure_open()
        except aioamqp.exceptions.AioamqpException:
            await self._close()
            raise

    async def startup(self, app: web.Application):
        pass  # Connections are created when attempting to publish

    async def shutdown(self, app: web.Application):
        await self._close()

    async def publish(self,
                      exchange: str,
                      routing: str,
                      message: Union[str, dict],
                      exchange_type: ExchangeType_ = 'topic'):
        """
        Publish a new event message.

        Connections are created automatically when calling `publish()`,
        and will attempt to reconnect if connection was lost.

        For more information on publishing AMQP messages,
        see https://www.rabbitmq.com/tutorials/tutorial-three-python.html

        Args:
            exchange (str):
                The AMQP message exchange to publish the message to.
                A new exchange will be created if it does not yet exist.

            routing (str):
                The routing identification with which the message should be published.
                Subscribers use routing information for fine-grained filtering.
                Routing can be expressed as a '.'-separated path.

            message (Union[str, dict]):
                The message body. It will be serialized before transmission.

            exchange_type (ExchangeType_, optional):
                When publishing to a previously undeclared exchange, it will be created.
                `exchange_type` defines how the exchange distributes messages between subscribers.
                The default is 'topic', and acceptable values are: 'topic', 'direct', or 'fanout'.

        Raises:
            aioamqp.exceptions.AioamqpException:
                * Failed to connect to AMQP host
                * Failed to send message
                * `exchange` already exists, but has a different `exchange_type`
        """
        try:
            await self._ensure_channel()
        except Exception:
            # If server has restarted since our last attempt, ensure channel will fail (old connection invalid)
            # Retry once to check whether a new connection can be made
            await self._ensure_channel()

        # json.dumps() also correctly handles strings
        data = json.dumps(message).encode()

        await self._channel.exchange_declare(
            exchange_name=exchange,
            type_name=exchange_type,
            auto_delete=True,
        )

        await self._channel.basic_publish(
            payload=data,
            exchange_name=exchange,
            routing_key=routing,
        )


##############################################################################
# Module functions
##############################################################################


def setup(app: web.Application):
    """Enables event listening / publishing.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    features.add(app, EventListener(app))
    features.add(app, EventPublisher(app))
    app.router.add_routes(routes)


def get_listener(app: web.Application) -> EventListener:
    """Gets registered EventListener.
    Requires setup() to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, EventListener)


def get_publisher(app: web.Application) -> EventPublisher:
    """Gets registered EventPublisher.
    Requires setup() to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, EventPublisher)


def subscribe(app: web.Application,
              exchange_name: str,
              routing: str,
              exchange_type: ExchangeType_ = 'topic',
              on_message: EvtCallbackType_ = None
              ) -> EventSubscription:
    """Adds a new event subscription to the listener.

    Actual queue declaration to the remote message server is done when connected.
    If the listener is not currently connected, it defers declaration.

    All existing subscriptions are redeclared on the remote if `EventListener`
    loses and recreates the connection.

    Shortcut for `EventListener.subscribe()`
    It requires setup() to have been called.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        exchange_name (str):
            Name of the AMQP exchange. Messages are always published to a specific exchange.

        routing (str):
            Filter messages passing through the exchange.
            A routing key is a '.'-separated string, and accepts '#' and '*' wildcards.

        exchange_type (ExchangeType_, optional):
            If the exchange does not yet exist, it will be created with this type.
            Default is `topic`, acceptable values are `topic`, `fanout`, or `direct`.

        on_message (EvtCallbackType_, optional):
            The function to be called when a new message is received.
            If `on_message` is none, it will default to logging the message.

    Returns:
        EventSubscription:
            The newly created subscription.
            This value can safely be discarded: EventListener keeps its own reference.
    """
    return get_listener(app).subscribe(exchange_name,
                                       routing,
                                       exchange_type,
                                       on_message)


async def publish(app: web.Application,
                  exchange: str,
                  routing: str,
                  message: Union[str, dict],
                  exchange_type: ExchangeType_ = 'topic'):
    """
    Publish a new event message.

    Connections are created automatically when calling `publish()`,
    and will attempt to reconnect if connection was lost.

    For more information on publishing AMQP messages,
    see https://www.rabbitmq.com/tutorials/tutorial-three-python.html

    Shortcut for `EventPublisher.publish()`.
    It requires setup() to have been called.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        exchange (str):
            The AMQP message exchange to publish the message to.
            A new exchange will be created if it does not yet exist.

        routing (str):
            The routing identification with which the message should be published.
            Subscribers use routing information for fine-grained filtering.
            Routing can be expressed as a '.'-separated path.

        message (Union[str, dict]):
            The message body. It will be serialized before transmission.

        exchange_type (ExchangeType_, optional):
            When publishing to a previously undeclared exchange, it will be created.
            `exchange_type` defines how the exchange distributes messages between subscribers.
            The default is 'topic', and acceptable values are: 'topic', 'direct', or 'fanout'.

    Raises:
        aioamqp.exceptions.AioamqpException:
            * Failed to connect to AMQP host
            * Failed to send message
            * `exchange` already exists, but has a different `exchange_type`
    """
    await get_publisher(app).publish(exchange,
                                     routing,
                                     message,
                                     exchange_type)


##############################################################################
# REST endpoints
##############################################################################

@routes.post('/_debug/publish')
async def post_publish(request):
    """
    ---
    tags:
    - Events
    summary: Publish event.
    description: Publish a new event message to the event bus.
    operationId: events.publish
    produces:
    - text/plain
    parameters:
    -
        in: body
        name: body
        description: Event message
        required: true
        schema:
            type: object
            properties:
                exchange:
                    type: string
                routing:
                    type: string
                message:
                    type: object
    """
    args = await request.json()
    await publish(request.app,
                  args['exchange'],
                  args['routing'],
                  args['message'])
    return web.Response()


@routes.post('/_debug/subscribe')
async def post_subscribe(request):
    """
    ---
    tags:
    - Events
    summary: Subscribe to events.
    operationId: events.subscribe
    produces:
    - text/plain
    parameters:
    -
        in: body
        name: body
        description: Event message
        required: true
        schema:
            type: object
            properties:
                exchange:
                    type: string
                routing:
                    type: string
    """
    args = await request.json()
    subscribe(request.app,
              args['exchange'],
              args['routing'])
    return web.Response()



"""
Tests brewblox_service.couchdb
"""

import pytest
from aiohttp import web
from aiohttp.client_exceptions import ClientResponseError

from brewblox_service import couchdb, http
from brewblox_service.couchdb import check_remote, read, write

TESTED = couchdb.__name__
SRV_URL = couchdb.COUCH_URL[len('http://'):]
DB_URL = '/sparkbase'
DOC_URL = '/sparkbase/sparkdoc'


@pytest.fixture
def app(app, mocker):
    mocker.patch(TESTED + '.DB_RETRY_INTERVAL_S', 0.01)
    http.setup(app)
    couchdb.setup(app)
    return app


async def test_check_remote(app, client, aresponses):
    for i in range(20):
        aresponses.add(SRV_URL, '/', 'HEAD', web.json_response({}, status=404))
    aresponses.add(SRV_URL, '/', 'HEAD')
    await check_remote(app)


async def test_read(app, client, aresponses):

    # Blank database
    aresponses.add(SRV_URL, '/', 'HEAD')
    aresponses.add(SRV_URL, DB_URL, 'PUT')
    aresponses.add(SRV_URL, DOC_URL, 'GET', web.json_response({}, status=404))
    aresponses.add(SRV_URL, DOC_URL, 'PUT', web.json_response({'rev': 'rev_read'}))
    assert await read(app, 'sparkbase', 'sparkdoc', [1, 2]) == ('rev_read', [1, 2])

    # Retry contact server, content in database
    for i in range(20):
        aresponses.add(SRV_URL, '/', 'HEAD', web.json_response({}, status=404))
    aresponses.add(SRV_URL, '/', 'HEAD')
    aresponses.add(SRV_URL, DB_URL, 'PUT', web.json_response({}, status=412))
    aresponses.add(SRV_URL, DOC_URL, 'GET', web.json_response({'_rev': 'rev_read', 'data': [2, 1]}))
    aresponses.add(SRV_URL, DOC_URL, 'PUT', web.json_response({}, status=409))
    assert await read(app, 'sparkbase', 'sparkdoc', []) == ('rev_read', [2, 1])


async def test_read_errors(app, client, aresponses):
    with pytest.raises(ClientResponseError):
        aresponses.add(SRV_URL, '/', 'HEAD')
        aresponses.add(SRV_URL, DB_URL, 'PUT', web.json_response({}, status=404))
        await read(app, 'sparkbase', 'sparkdoc', [])

    with pytest.raises(ClientResponseError):
        aresponses.add(SRV_URL, '/', 'HEAD')
        aresponses.add(SRV_URL, DB_URL, 'PUT')
        aresponses.add(SRV_URL, DOC_URL, 'PUT', web.json_response({}, status=404))  # unexpected
        aresponses.add(SRV_URL, DOC_URL, 'GET', web.json_response({}, status=404))
        await read(app, 'sparkbase', 'sparkdoc', [])

    with pytest.raises(ClientResponseError):
        aresponses.add(SRV_URL, '/', 'HEAD')
        aresponses.add(SRV_URL, DB_URL, 'PUT')
        aresponses.add(SRV_URL, DOC_URL, 'PUT', web.json_response({}, status=412))
        aresponses.add(SRV_URL, DOC_URL, 'GET', web.json_response({}, status=500))  # unexpected
        await read(app, 'sparkbase', 'sparkdoc', [])

    with pytest.raises(ValueError):
        aresponses.add(SRV_URL, '/', 'HEAD')
        aresponses.add(SRV_URL, DB_URL, 'PUT')
        # Either get or put must return an ok value
        aresponses.add(SRV_URL, DOC_URL, 'PUT', web.json_response({}, status=409))
        aresponses.add(SRV_URL, DOC_URL, 'GET', web.json_response({}, status=404))
        await read(app, 'sparkbase', 'sparkdoc', [])


async def test_write(app, client, aresponses):
    aresponses.add(
        SRV_URL, f'{DOC_URL}?rev=revy', 'PUT',
        web.json_response({'rev': 'rev_write'}), match_querystring=True)
    assert await write(app, 'sparkbase', 'sparkdoc', 'revy', [1, 2]) == 'rev_write'



"""
Tests brewblox_service.repeater
"""

import asyncio
from unittest.mock import Mock

import pytest

from brewblox_service import features, repeater, scheduler

TESTED = repeater.__name__


class RepeaterDummy(repeater.RepeaterFeature):

    def __init__(self, app):
        self.prepare_mock = Mock()
        self.run_mock = Mock()
        self.interval = 0.01
        super().__init__(app)

    async def prepare(self):
        self.prepare_mock()

    async def run(self):
        await asyncio.sleep(self.interval)
        self.run_mock()


class ResumeDummy(RepeaterDummy):

    async def run(self):
        await asyncio.sleep(self.interval)
        self.run_mock()
        if self.run_mock.call_count == 1:
            raise RuntimeError()


@pytest.fixture
async def app(app):
    scheduler.setup(app)

    dummy = RepeaterDummy(app)
    features.add(app, dummy, 'dummy')

    prep_cancel = RepeaterDummy(app)
    prep_cancel.prepare_mock.side_effect = repeater.RepeaterCancelled
    features.add(app, prep_cancel, 'prep_cancel')

    prep_error = RepeaterDummy(app)
    prep_error.prepare_mock.side_effect = RuntimeError
    features.add(app, prep_error, 'prep_error')

    run_cancel = RepeaterDummy(app)
    run_cancel.run_mock.side_effect = repeater.RepeaterCancelled
    features.add(app, run_cancel, 'run_cancel')

    run_error = RepeaterDummy(app)
    run_error.run_mock.side_effect = RuntimeError
    features.add(app, run_error, 'run_error')

    run_resume = ResumeDummy(app)
    features.add(app, run_resume, 'run_resume')

    return app


async def test_dummies(app, client):
    dummy = features.get(app, key='dummy')
    prep_cancel = features.get(app, key='prep_cancel')
    prep_error = features.get(app, key='prep_error')
    run_cancel = features.get(app, key='run_cancel')
    run_error = features.get(app, key='run_error')
    run_resume = features.get(app, key='run_resume')

    await asyncio.sleep(0.1)

    assert dummy.active
    assert dummy.prepare_mock.call_count == 1
    assert dummy.run_mock.call_count > 1

    assert not prep_cancel.active
    assert prep_cancel.prepare_mock.call_count == 1
    assert prep_cancel.run_mock.call_count == 0

    assert not prep_error.active
    assert prep_error.prepare_mock.call_count == 1
    assert prep_error.run_mock.call_count == 0

    assert not run_cancel.active
    assert run_cancel.prepare_mock.call_count == 1
    assert run_cancel.run_mock.call_count == 1

    assert run_error.active
    assert run_error.prepare_mock.call_count == 1
    assert run_error.run_mock.call_count > 1

    assert run_resume.active
    assert run_resume.prepare_mock.call_count == 1
    assert run_resume.run_mock.call_count > 1



"""
Cross Origin Resource Sharing (CORS) headers must be present when using multiple backends.
This middleware automatically returns OPTIONS requests, and appends other requests with the correct headers.
"""

from aiohttp import hdrs, web, web_exceptions

from brewblox_service import brewblox_logger, strex

LOGGER = brewblox_logger(__name__)


def enable_cors(app: web.Application):
    app.middlewares.append(cors_middleware)


def set_cors_headers(request, response):
    response.headers['Access-Control-Allow-Origin'] = '*'
    response.headers['Access-Control-Allow-Methods'] =\
        request.headers.get('Access-Control-Request-Method', ','.join(hdrs.METH_ALL))
    response.headers['Access-Control-Allow-Headers'] =\
        request.headers.get('Access-Control-Request-Headers', '*')
    response.headers['Access-Control-Allow-Credentials'] = 'true'
    return response


@web.middleware
async def cors_middleware(request: web.Request, handler) -> web.Response:
    # preflight requests
    if request.method == 'OPTIONS':
        return set_cors_headers(request, web.Response())
    else:
        try:
            response = await handler(request)
        except web_exceptions.HTTPError as ex:
            response = ex
        except Exception as ex:
            response = web_exceptions.HTTPInternalServerError(reason=strex(ex))

        return set_cors_headers(request, response)



"""
Background task scheduling.
"""

import asyncio
from contextlib import suppress
from typing import Any, Coroutine, Optional, Set

from aiohttp import web

from brewblox_service import brewblox_logger, features

CLEANUP_INTERVAL_S = 300


LOGGER = brewblox_logger(__name__)


class TaskScheduler(features.ServiceFeature):

    def __init__(self, app: web.Application):
        super().__init__(app)
        self._tasks: Set[asyncio.Task] = set()

    async def startup(self, *_):
        await self.create(self._cleanup())

    async def shutdown(self, *_):
        for task in self._tasks:
            task.cancel()
        await asyncio.wait(self._tasks)
        self._tasks.clear()

    async def _cleanup(self):
        """
        Periodically removes completed tasks from the collection,
        allowing fire-and-forget tasks to be garbage collected.

        This does not delete the task object, it merely removes the reference in the scheduler.
        """
        while True:
            await asyncio.sleep(CLEANUP_INTERVAL_S)
            self._tasks = set(t for t in self._tasks if not t.done())

    async def create(self,
                     coro: Coroutine,
                     name: str = None,
                     ) -> asyncio.Task:
        task = asyncio.create_task(coro, name=name)
        LOGGER.debug(f'Scheduled {task}')
        self._tasks.add(task)
        return task

    async def cancel(self,
                     task: asyncio.Task,
                     wait_for: bool = True,
                     ) -> Optional[Any]:
        if task is None:
            return

        task.cancel()

        with suppress(KeyError):
            self._tasks.remove(task)

        retv = None

        with suppress(Exception, asyncio.CancelledError):
            if wait_for:
                retv = await task

        LOGGER.debug(f'Cancelled: {task}')
        return retv


def setup(app: web.Application):
    """Registers scheduler functionality with Application.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    features.add(app, TaskScheduler(app))


def get_scheduler(app: web.Application) -> TaskScheduler:
    """Gets the default TaskScheduler object

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, TaskScheduler)


async def create(app: web.Application,
                 coro: Coroutine,
                 name: str = None
                 ) -> asyncio.Task:
    """
    Starts execution of a coroutine.

    The created asyncio.Task is returned, and added to managed tasks.
    The scheduler guarantees that it is cancelled during application shutdown,
    regardless of whether it was already cancelled manually.

    Shortcut for `TaskScheduler.create(coro)`
    Requires setup() to have been called.

    Args:
        coro (Coroutine):
            The coroutine to be wrapped in a task, and executed.
        name (str):
            Custom name for the created task.

    Returns:
        asyncio.Task: An awaitable Task object.
            During Aiohttp shutdown, the scheduler will attempt to cancel and await this task.
            The task can be safely cancelled manually, or using `TaskScheduler.cancel(task)`.

    Example:

        import asyncio
        from datetime import datetime
        from brewblox_service import scheduler, service


        async def current_time(interval):
            while True:
                await asyncio.sleep(interval)
                print(datetime.now())


        async def start(app):
            await scheduler.create(app, current_time(interval=2))


        app = service.create_app(default_name='example')

        scheduler.setup(app)
        app.on_startup.append(start)

        service.furnish(app)
        service.run(app)
    """
    return await get_scheduler(app).create(coro, name=name)


async def cancel(app: web.Application,
                 task: asyncio.Task,
                 wait_for: bool = True,
                 ) -> Optional[Any]:
    """
    Cancels and waits for an `asyncio.Task` to finish.
    Removes it from the collection of managed tasks.

    Shortcut for `TaskScheduler.cancel(task)`.
    Requires setup() to have been called.

    Args:
        task (asyncio.Task):
            The to be cancelled task.
            It is not required that the task was was created with `TaskScheduler.create()`.

        wait_for (bool, optional):
            Whether to wait for the task to finish execution.
            If falsey, this function returns immediately after cancelling the task.

    Returns:
        Any: The return value of `task`. None if `wait_for` is falsey.

    Example:

        import asyncio
        from datetime import datetime
        from brewblox_service import scheduler, service

        async def current_time(interval):
            while True:
                await asyncio.sleep(interval)
                print(datetime.now())


        async def stop_after(app, task, duration):
            await asyncio.sleep(duration)
            await scheduler.cancel(app, task)
            print('stopped!')


        async def start(app):
            # Start first task
            task = await scheduler.create(app, current_time(interval=2))

            # Start second task to stop the first
            await scheduler.create(app, stop_after(app, task, duration=10))


        app = service.create_app(default_name='example')

        scheduler.setup(app)
        app.on_startup.append(start)

        service.furnish(app)
        service.run(app)
    """
    return await get_scheduler(app).cancel(task, wait_for=wait_for)
